/*
 * Copyright (c) 2005 The University of Notre Dame. All Rights Reserved.
 *
 * The University of Notre Dame grants you ("Licensee") a
 * non-exclusive, royalty free, license to use, modify and
 * redistribute this software in source and binary code form, provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. All express or implied conditions, representations and
 * warranties, including any implied warranty of merchantability,
 * fitness for a particular purpose or non-infringement, are hereby
 * excluded.  The University of Notre Dame and its licensors shall not
 * be liable for any damages suffered by licensee as a result of
 * using, modifying or distributing the software or its
 * derivatives. In no event will the University of Notre Dame or its
 * licensors be liable for any lost revenue, profit or data, or for
 * direct, indirect, special, consequential, incidental or punitive
 * damages, however caused and regardless of the theory of liability,
 * arising out of the use of or inability to use software, even if the
 * University of Notre Dame has been advised of the possibility of
 * such damages.
 *
 * SUPPORT OPEN SCIENCE!  If you use OpenMD or its source code in your
 * research, please cite the appropriate papers when you publish your
 * work.  Good starting points are:
 *                                                                      
 * [1]  Meineke, et al., J. Comp. Chem. 26, 252-271 (2005).             
 * [2]  Fennell & Gezelter, J. Chem. Phys. 124, 234104 (2006).          
 * [3]  Sun, Lin & Gezelter, J. Chem. Phys. 128, 234107 (2008).          
 * [4]  Kuang & Gezelter,  J. Chem. Phys. 133, 164101 (2010).
 * [5]  Vardeman, Stocker & Gezelter, J. Chem. Theory Comput. 7, 834 (2011).
 * [6]  Kuang & Gezelter, Mol. Phys., 110, 691-701 (2012).
 */

#ifndef APPLICATIONS_STATICPROPS_FIELD_HPP
#define APPLICATIONS_STATICPROPS_FIELD_HPP

#include "selection/SelectionEvaluator.hpp"
#include "selection/SelectionManager.hpp"
#include "applications/staticProps/StaticAnalyser.hpp"
#include "math/Vector3.hpp"

namespace OpenMD {


  template<class T>
  class Field : public StaticAnalyser{
    
  public:
    Field(SimInfo* info, const std::string& filename, 
	  const std::string& sele, RealType voxelSize);
    
    ~Field(); // default deconstructor
    
    virtual void process();
    virtual void processFrame(int frame);
    virtual void postProcess();
    virtual T getValue(StuntDouble* sd) = 0;
    virtual void writeField();
    virtual std::string writeValue(T v);
    virtual void writeVisualizationScript();
    
  protected:
    RealType getDensity(RealType dist, RealType sigma, RealType rcut);
    
    Snapshot* snap_;
    int nProcessed_;
    string selectionScript_;
    SelectionManager seleMan_;
    SelectionEvaluator evaluator_;
    bool usePeriodicBoundaryConditions_;
    RealType rcut_;
    RealType reffective_;
    
    RealType voxelSize_;
    Vector3i nBins_;
    
    std::vector<std::vector<std::vector<RealType> > > dens_;
    std::vector<std::vector<std::vector<T > > > field_;
  };

  class DensityField : public Field<RealType> {
  public:
    DensityField(SimInfo* info, const std::string& filename,
                 const std::string& sele1, RealType voxelSize);
    ~DensityField();
    
    virtual RealType getValue(StuntDouble* sd);
  };
  
  
  class ChargeField : public Field<RealType> {
  public:
    ChargeField(SimInfo* info, const std::string& filename,
                       const std::string& sele1, RealType voxelSize);
    virtual RealType getValue(StuntDouble* sd);
  };
  
  
  class VelocityField : public Field<Vector3d> {
  public:
    VelocityField(SimInfo* info, const std::string& filename,
			const std::string& sele1, RealType voxelSize);
    virtual Vector3d getValue(StuntDouble* sd);
  };
  
  
  class DipoleField : public Field<Vector3d> {
  public:
    DipoleField(SimInfo* info, const std::string& filename,
                const std::string& sele1, RealType voxelSize);
    virtual Vector3d getValue(StuntDouble* sd);
  };
  
}
#endif

