/*
 * Copyright (c) 2005 The University of Notre Dame. All Rights Reserved.
 *
 * The University of Notre Dame grants you ("Licensee") a
 * non-exclusive, royalty free, license to use, modify and
 * redistribute this software in source and binary code form, provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. All express or implied conditions, representations and
 * warranties, including any implied warranty of merchantability,
 * fitness for a particular purpose or non-infringement, are hereby
 * excluded.  The University of Notre Dame and its licensors shall not
 * be liable for any damages suffered by licensee as a result of
 * using, modifying or distributing the software or its
 * derivatives. In no event will the University of Notre Dame or its
 * licensors be liable for any lost revenue, profit or data, or for
 * direct, indirect, special, consequential, incidental or punitive
 * damages, however caused and regardless of the theory of liability,
 * arising out of the use of or inability to use software, even if the
 * University of Notre Dame has been advised of the possibility of
 * such damages.
 *
 * SUPPORT OPEN SCIENCE!  If you use OpenMD or its source code in your
 * research, please cite the appropriate papers when you publish your
 * work.  Good starting points are:
 *
 * [1]  Meineke, et al., J. Comp. Chem. 26, 252-271 (2005).
 * [2]  Fennell & Gezelter, J. Chem. Phys. 124, 234104 (2006).
 * [3]  Sun, Lin & Gezelter, J. Chem. Phys. 128, 234107 (2008).
 * [4] Kuang & Gezelter,  J. Chem. Phys. 133, 164101 (2010).
 */


#include <algorithm>
#include <fstream>
#include "applications/staticProps/RNEMDStats.hpp"
#include "primitives/Molecule.hpp"
#include "utils/Constants.hpp"
#include "types/FixedChargeAdapter.hpp"
#include "types/FluctuatingChargeAdapter.hpp"

namespace OpenMD {

  RNEMDZ::RNEMDZ(SimInfo* info, const std::string& filename,
                 const std::string& sele, int nzbins, int axis)
    : SlabStatistics(info, filename, sele, nzbins, axis), axis_(axis) {


    setOutputName(getPrefix(filename) + ".rnemdZ");

    temperature = new OutputData;
    temperature->units =  "K";
    temperature->title =  "Temperature";
    temperature->dataType = odtReal;
    temperature->dataHandling = odhAverage;
    temperature->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      temperature->accumulator.push_back( new Accumulator() );
    data_.push_back(temperature);

    velocity = new OutputData;
    velocity->units = "angstroms/fs";
    velocity->title =  "Velocity";
    velocity->dataType = odtVector3;
    velocity->dataHandling = odhAverage;
    velocity->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      velocity->accumulator.push_back( new VectorAccumulator() );
    data_.push_back(velocity);

    density = new OutputData;
    density->units =  "g cm^-3";
    density->title =  "Density";
    density->dataType = odtReal;
    density->dataHandling = odhAverage;
    density->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      density->accumulator.push_back( new Accumulator() );
    data_.push_back(density);

    charge = new OutputData;
    charge->units =  "e";
    charge->title =  "Charge";
    charge->dataType = odtReal;
    charge->dataHandling = odhAverage;
    charge->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      charge->accumulator.push_back( new Accumulator() );
    data_.push_back(charge);


    chargeVelocity = new OutputData;
    chargeVelocity->units =  "e/fs";
    chargeVelocity->title =  "Charge_Velocity";
    chargeVelocity->dataType = odtReal;
    chargeVelocity->dataHandling = odhAverage;
    chargeVelocity->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      chargeVelocity->accumulator.push_back( new Accumulator() );
    data_.push_back(chargeVelocity);
  }

  void RNEMDZ::processFrame(int istep) {
    RealType z;

    hmat_ = currentSnapshot_->getHmat();
    for (unsigned int i = 0; i < nBins_; i++) {
      z = (((RealType)i + 0.5) / (RealType)nBins_) * hmat_(axis_,axis_);
      dynamic_cast<Accumulator*>(z_->accumulator[i])->add(z);
    }
    volume_ = currentSnapshot_->getVolume();

    StuntDouble* sd;
    int i;

    vector<RealType> binMass(nBins_, 0.0);
    vector<Vector3d> binP(nBins_, V3Zero);
    vector<RealType> binCharge(nBins_, 0.0);
    vector<RealType> binChargeVelocity(nBins_, 0.0);
    vector<RealType> binKE(nBins_, 0.0);
    vector<unsigned int> binDof(nBins_, 0);

    if (evaluator_.isDynamic()) {
      seleMan_.setSelectionSet(evaluator_.evaluate());
    }

    // loop over the selected atoms:

    for (sd = seleMan_.beginSelected(i); sd != NULL;
         sd = seleMan_.nextSelected(i)) {

      // figure out where that object is:
      Vector3d pos = sd->getPos();
      Vector3d vel = sd->getVel();
      RealType m = sd->getMass();

      int bin = getBin(pos);

      binMass[bin] += m;
      binP[bin] += m * vel;
      binKE[bin] += 0.5 * (m * vel.lengthSquare());
      binDof[bin] += 3;

      if (sd->isDirectional()) {
        Vector3d angMom = sd->getJ();
        Mat3x3d I = sd->getI();
        if (sd->isLinear()) {
          int i = sd->linearAxis();
          int j = (i + 1) % 3;
          int k = (i + 2) % 3;
          binKE[bin] += 0.5 * (angMom[j] * angMom[j] / I(j, j) +
                               angMom[k] * angMom[k] / I(k, k));
          binDof[bin] += 2;
        } else {
          binKE[bin] += 0.5 * (angMom[0] * angMom[0] / I(0, 0) +
                               angMom[1] * angMom[1] / I(1, 1) +
                               angMom[2] * angMom[2] / I(2, 2));
          binDof[bin] += 3;
        }
      }

      RealType q = 0.0;
      RealType w = 0.0;
      if(sd->isAtom()){

        Atom* atom = static_cast<Atom*>(sd);

        AtomType* atomType = atom->getAtomType();

        FixedChargeAdapter fca = FixedChargeAdapter(atomType);
        if ( fca.isFixedCharge() ) {
          q += fca.getCharge();
        }

        FluctuatingChargeAdapter fqa = FluctuatingChargeAdapter(atomType);
        if ( fqa.isFluctuatingCharge() ) {
          q += atom->getFlucQPos();
          w += sd->getFlucQVel();
        }

        binCharge[bin] += q;
        binChargeVelocity[bin] += w;


      }
    }


    for (unsigned int i = 0; i < nBins_; i++) {

      if (binDof[i] > 0) {
        RealType temp = 2.0 * binKE[i] / (binDof[i] * Constants::kb *
                                          Constants::energyConvert);
        RealType den = binMass[i] * nBins_ * Constants::densityConvert
          / volume_;
        Vector3d vel = binP[i] / binMass[i];

        dynamic_cast<Accumulator *>(temperature->accumulator[i])->add(temp);
        dynamic_cast<VectorAccumulator *>(velocity->accumulator[i])->add(vel);
        dynamic_cast<Accumulator *>(density->accumulator[i])->add(den);
        dynamic_cast<Accumulator *>(charge->accumulator[i])->add(binCharge[i]);
        dynamic_cast<Accumulator *>(chargeVelocity->accumulator[i])->add(binChargeVelocity[i]);
        dynamic_cast<Accumulator *>(counts_->accumulator[i])->add(1);
      }
    }
  }

  void RNEMDZ::processStuntDouble(StuntDouble* sd, int bin) {
  }

  RNEMDR::RNEMDR(SimInfo* info, const std::string& filename,
                 const std::string& sele, int nrbins)
    : ShellStatistics(info, filename, sele, nrbins) {


    setOutputName(getPrefix(filename) + ".rnemdR");

    temperature = new OutputData;
    temperature->units =  "K";
    temperature->title =  "Temperature";
    temperature->dataType = odtReal;
    temperature->dataHandling = odhAverage;
    temperature->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      temperature->accumulator.push_back( new Accumulator() );
    data_.push_back(temperature);

    angularVelocity = new OutputData;
    angularVelocity->units = "angstroms^2/fs";
    angularVelocity->title =  "Angular Velocity";
    angularVelocity->dataType = odtVector3;
    angularVelocity->dataHandling = odhAverage;
    angularVelocity->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      angularVelocity->accumulator.push_back( new VectorAccumulator() );
    data_.push_back(angularVelocity);

    density = new OutputData;
    density->units =  "g cm^-3";
    density->title =  "Density";
    density->dataType = odtReal;
    density->dataHandling = odhAverage;
    density->accumulator.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++)
      density->accumulator.push_back( new Accumulator() );
    data_.push_back(density);
  }


  void RNEMDR::processFrame(int istep) {

    StuntDouble* sd;
    int i;

    vector<RealType> binMass(nBins_, 0.0);
    vector<Mat3x3d>  binI(nBins_);
    vector<Vector3d> binL(nBins_, V3Zero);
    vector<RealType> binKE(nBins_, 0.0);
    vector<int> binDof(nBins_, 0);

    if (evaluator_.isDynamic()) {
      seleMan_.setSelectionSet(evaluator_.evaluate());
    }

    // loop over the selected atoms:

    for (sd = seleMan_.beginSelected(i); sd != NULL;
         sd = seleMan_.nextSelected(i)) {

      // figure out where that object is:
      int bin = getBin( sd->getPos() );

      if (bin >= 0 && bin < int(nBins_))  {

        Vector3d rPos = sd->getPos() - coordinateOrigin_;
        Vector3d vel = sd->getVel();
        RealType m = sd->getMass();
        Vector3d L = m * cross(rPos, vel);
        Mat3x3d I(0.0);
        I = outProduct(rPos, rPos) * m;
        RealType r2 = rPos.lengthSquare();
        I(0, 0) += m * r2;
        I(1, 1) += m * r2;
        I(2, 2) += m * r2;

        binMass[bin] += m;
        binI[bin] += I;
        binL[bin] += L;
        binKE[bin] += 0.5 * (m * vel.lengthSquare());
        binDof[bin] += 3;

        if (sd->isDirectional()) {
          Vector3d angMom = sd->getJ();
          Mat3x3d Ia = sd->getI();
          if (sd->isLinear()) {
            int i = sd->linearAxis();
            int j = (i + 1) % 3;
            int k = (i + 2) % 3;
            binKE[bin] += 0.5 * (angMom[j] * angMom[j] / Ia(j, j) +
                                 angMom[k] * angMom[k] / Ia(k, k));
            binDof[bin] += 2;
          } else {
            binKE[bin] += 0.5 * (angMom[0] * angMom[0] / Ia(0, 0) +
                                 angMom[1] * angMom[1] / Ia(1, 1) +
                                 angMom[2] * angMom[2] / Ia(2, 2));
            binDof[bin] += 3;
          }
        }
      }
    }

    for (unsigned int i = 0; i < nBins_; i++) {
      RealType rinner = (RealType)i * binWidth_;
      RealType router = (RealType)(i+1) * binWidth_;
      if (binDof[i] > 0) {
        RealType temp = 2.0 * binKE[i] / (binDof[i] * Constants::kb *
                                          Constants::energyConvert);
        RealType den = binMass[i] * 3.0 * Constants::densityConvert
          / (4.0 * Constants::PI * (pow(router,3) - pow(rinner,3)));

        Vector3d omega = binI[i].inverse() * binL[i];

        dynamic_cast<Accumulator *>(temperature->accumulator[i])->add(temp);
        dynamic_cast<VectorAccumulator *>(angularVelocity->accumulator[i])->add(omega);
        dynamic_cast<Accumulator *>(density->accumulator[i])->add(den);
        dynamic_cast<Accumulator *>(counts_->accumulator[i])->add(1);
      }
    }
  }


  void RNEMDR::processStuntDouble(StuntDouble* sd, int bin) {
  }

  RNEMDRTheta::RNEMDRTheta(SimInfo* info, const std::string& filename,
                           const std::string& sele, int nrbins, int nangleBins)
    : ShellStatistics(info, filename, sele, nrbins), nAngleBins_(nangleBins) {

    Globals* simParams = info->getSimParams();
    RNEMDParameters* rnemdParams = simParams->getRNEMDParameters();
    bool hasAngularMomentumFluxVector = rnemdParams->haveAngularMomentumFluxVector();

    if (hasAngularMomentumFluxVector) {
      std::vector<RealType> amf = rnemdParams->getAngularMomentumFluxVector();
      if (amf.size() != 3) {
        sprintf(painCave.errMsg,
                "RNEMDRTheta: Incorrect number of parameters specified for angularMomentumFluxVector.\n"
                "\tthere should be 3 parameters, but %lu were specified.\n",
                amf.size());
        painCave.isFatal = 1;
        simError();
      }
      fluxVector_.x() = amf[0];
      fluxVector_.y() = amf[1];
      fluxVector_.z() = amf[2];
    } else {

      std::string fluxStr = rnemdParams->getFluxType();
      if (fluxStr.find("Lx") != std::string::npos) {
        fluxVector_ = V3X;
      } else if (fluxStr.find("Ly") != std::string::npos) {
        fluxVector_ = V3Y;
      } else {
        fluxVector_ = V3Z;
      }
    }

    fluxVector_.normalize();

    setOutputName(getPrefix(filename) + ".rnemdRTheta");

    angularVelocity = new OutputData;
    angularVelocity->units = "angstroms^2/fs";
    angularVelocity->title =  "Angular Velocity";
    angularVelocity->dataType = odtArray2d;
    angularVelocity->dataHandling = odhAverage;
    angularVelocity->accumulatorArray2d.reserve(nBins_);
    for (unsigned int i = 0; i < nBins_; i++) {
      angularVelocity->accumulatorArray2d[i].reserve(nAngleBins_);
      for (int j = 0 ; j < nAngleBins_; j++) {
        angularVelocity->accumulatorArray2d[i][j] = new Accumulator();
      }
    }
    data_.push_back(angularVelocity);

  }


  std::pair<int,int> RNEMDRTheta::getBins(Vector3d pos) {
    std::pair<int,int> result;

    Vector3d rPos = pos - coordinateOrigin_;
    RealType cosAngle= dot(rPos, fluxVector_) / rPos.length();

    result.first = int(rPos.length() / binWidth_);
    result.second = int( (nAngleBins_ - 1) * 0.5 * (cosAngle + 1.0) );
    return result;
  }

  void RNEMDRTheta::processStuntDouble(StuntDouble* sd, int bin) {
  }

  void RNEMDRTheta::processFrame(int istep) {

    StuntDouble* sd;
    int i;

    vector<vector<Mat3x3d> >  binI;
    vector<vector<Vector3d> > binL;
    vector<vector<int> > binCount;

    if (evaluator_.isDynamic()) {
      seleMan_.setSelectionSet(evaluator_.evaluate());
    }

    // loop over the selected atoms:

    for (sd = seleMan_.beginSelected(i); sd != NULL;
         sd = seleMan_.nextSelected(i)) {

      // figure out where that object is:
      std::pair<int,int> bins = getBins( sd->getPos() );

      if (bins.first >= 0 && bins.first < int(nBins_))  {
        if (bins.second >= 0 && bins.second < nAngleBins_) {

          Vector3d rPos = sd->getPos() - coordinateOrigin_;
          Vector3d vel = sd->getVel();
          RealType m = sd->getMass();
          Vector3d L = m * cross(rPos, vel);
          Mat3x3d I(0.0);
          I = outProduct(rPos, rPos) * m;
          RealType r2 = rPos.lengthSquare();
          I(0, 0) += m * r2;
          I(1, 1) += m * r2;
          I(2, 2) += m * r2;

          binI[bins.first][bins.second] += I;
          binL[bins.first][bins.second] += L;
          binCount[bins.first][bins.second]++;
        }
      }
    }


    for (unsigned int i = 0; i < nBins_; i++) {
      for (int j = 0; j < nAngleBins_; j++) {

        if (binCount[i][j] > 0) {
          Vector3d omega = binI[i][j].inverse() * binL[i][j];
          RealType omegaProj = dot(omega, fluxVector_);

          dynamic_cast<Accumulator *>(angularVelocity->accumulatorArray2d[i][j])->add(omegaProj);
        }
      }
    }
  }

  void RNEMDRTheta::writeOutput() {

    vector<OutputData*>::iterator i;
    OutputData* outputData;

    ofstream outStream(outputFilename_.c_str());
    if (outStream.is_open()) {

      //write title
      outStream << "# SPATIAL STATISTICS\n";
      outStream << "#";

      for(outputData = beginOutputData(i); outputData;
          outputData = nextOutputData(i)) {
        outStream << "\t" << outputData->title <<
          "(" << outputData->units << ")";
        // add some extra tabs for column alignment
        if (outputData->dataType == odtVector3) outStream << "\t\t";
      }

      outStream << std::endl;

      outStream.precision(8);

      for (unsigned int j = 0; j < nBins_; j++) {

        int counts = counts_->accumulator[j]->count();

        if (counts > 0) {
          for(outputData = beginOutputData(i); outputData;
              outputData = nextOutputData(i)) {

            int n = outputData->accumulator[j]->count();
            if (n != 0) {
              writeData( outStream, outputData, j );
            }
          }
          outStream << std::endl;
        }
      }
    }
  }
}
