/*
 * Copyright (c) 2005 The University of Notre Dame. All Rights Reserved.
 *
 * The University of Notre Dame grants you ("Licensee") a
 * non-exclusive, royalty free, license to use, modify and
 * redistribute this software in source and binary code form, provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. All express or implied conditions, representations and
 * warranties, including any implied warranty of merchantability,
 * fitness for a particular purpose or non-infringement, are hereby
 * excluded.  The University of Notre Dame and its licensors shall not
 * be liable for any damages suffered by licensee as a result of
 * using, modifying or distributing the software or its
 * derivatives. In no event will the University of Notre Dame or its
 * licensors be liable for any lost revenue, profit or data, or for
 * direct, indirect, special, consequential, incidental or punitive
 * damages, however caused and regardless of the theory of liability,
 * arising out of the use of or inability to use software, even if the
 * University of Notre Dame has been advised of the possibility of
 * such damages.
 *
 * SUPPORT OPEN SCIENCE!  If you use OpenMD or its source code in your
 * research, please cite the appropriate papers when you publish your
 * work.  Good starting points are:
 *
 * [1]  Meineke, et al., J. Comp. Chem. 26, 252-271 (2005).
 * [2]  Fennell & Gezelter, J. Chem. Phys. 124, 234104 (2006).
 * [3]  Sun, Lin & Gezelter, J. Chem. Phys. 128, 234107 (2008).
 * [4]  Kuang & Gezelter,  J. Chem. Phys. 133, 164101 (2010).
 * [5]  Vardeman, Stocker & Gezelter, J. Chem. Theory Comput. 7, 834 (2011).
 */


/*! \file perturbations/MagneticField.hpp
    \brief Uniform Magnetic Field perturbation
*/

#ifndef PERTURBATIONS_MAGNETICFIELD_HPP
#define PERTURBATIONS_MAGNETICFIELD_HPP

#include "perturbations/Perturbation.hpp"
#include "brains/SimInfo.hpp"

namespace OpenMD {

  //! Applies a uniform (vector) magnetic field to the system
  /*! The field is applied as an external perturbation.  The user specifies

  \code{.unparsed}
    magneticField = (a, b, c);
  \endcode

    in the .omd file where the values of a, b, and c are in units of Tesla

    When a charge moves in a magnetic field, there is no work done, so
    there is no change in potential.

    The field itself is

    \f$ \mathbf{B} = \left( \begin{array}{c} a \\ b \\ c \end{array}
    \right) \f$

   The external magnetic field applies a force on moving charged
   atoms, \f$ \mathbf{F} = C \mathbf{v} \times \mathbf{B} \f$ and a
   torque on moving (and rotating) dipoles, \f$ \mathbf{\tau} =
   \mathbf{D} \times ( \mathbf{v} \times \mathbf{B} ) +
   (\mathbf{\omega} \times \mathbf{D} ) \times \mathbf{B} \f$

  */
  class MagneticField : public Perturbation {

  public:
    MagneticField(SimInfo* info);

  protected:
    virtual void initialize();
    virtual void applyPerturbation();

  private:
    bool initialized;
    bool doMagneticField;
    Globals* simParams;
    SimInfo* info_;
    Vector3d MF;
  };

} //end namespace OpenMD
#endif
